import logging
import astropy.units as u
import numpy as np


from stixcore.imaging.em import em
from stixcore.calibration.visibility import calibrate_visibility, create_visibility, \
    correct_phase_projection
from stixpy.science import ScienceData
from xrayvision.clean import vis_clean
from xrayvision.imaging import vis_to_image
from xrayvision.mem import mem
from xrayvision.visibility import Visibility

logger = logging.getLogger(__name__)
logger.setLevel('DEBUG')

cpd = ScienceData.from_fits('/Users/shane/Downloads/'
                            'solo_L1_stix-sci-xray-cpd_20200607T213709-20200607T215208'
                            '_V01_1178428688-49155.fits')


time_range = ['2020-06-07T21:39:00', '2020-06-07T21:42:49']
energy_range = [6, 10]
flare_xy = [-1625, -700] * u.arcsec

# TODO rename doesn't really create a complex vis just the real and imag parts
vis = create_visibility(cpd, time_range=time_range,
                        energy_range=energy_range, phase_center=flare_xy, no_shadowing=False)


# TODO rename
cal_vis = calibrate_visibility(vis)

cal_vis = correct_phase_projection(cal_vis, flare_xy)

# order by sub-collimator e.g 10a, 10b, 10c, 9a, 9b, 9c ....
isc_10_3 = [3, 20, 22, 16, 14, 32, 21, 26, 4, 24, 8, 28, 15, 27, 31, 6, 30, 2, 25, 5, 23, 7, 29, 1]
idx = np.searchsorted(cal_vis.isc, isc_10_3)

imsize = [129, 129]*u.pixel  # number of pixels of the map to reconstruct
pixel = [2, 2]*u.arcsec   # pixel size in aresec

stix_vis = Visibility(vis=cal_vis.obsvis[idx], u=cal_vis.u[idx], v=cal_vis.v[idx])
skeys = stix_vis.__dict__.keys()
for k, v in cal_vis.__dict__.items():
    if k not in skeys:
        setattr(stix_vis, k, v[idx])

# back projection (inverse transform) natural weighting
bp_nat = vis_to_image(stix_vis, imsize, pixel_size=pixel)

# back project (inverse transform) uniform weighting
bp_uni = vis_to_image(stix_vis, imsize, pixel_size=pixel, natural=False)

# Clean
# niter = 60  # number of iterations
# gain = 0.1  # gain used in each clean iteration
# beam_width = 20. * u.arcsec
# clean_map, model_map, resid_map = vis_clean(stix_vis, imsize, pixel=pixel, gain=gain, niter=niter,
#                                             clean_beam_width=20*u.arcsec)
#
# mem_map = mem(stix_vis, shape=imsize, pixel=pixel)

vis = create_visibility(cpd, time_range=time_range, energy_range=energy_range, phase_center=flare_xy)

em_map = em(vis['rate'], stix_vis, shape=imsize, pixel_size=pixel, flare_xy=flare_xy, idx=idx)
print(em_map)



from unittest.mock import patch

import astropy.units as u
import numpy as np
from astropy.time.core import Time
from sunpy.data import sample
from sunpy.map import Map

from stixpy.vis.map_reprojection import get_solo_position, reproject_map


@patch('sunpy.map.GenericMap')
def test_get_solo_position(mock_map):
    mock_map.date = Time('2021-04-17T00:00')
    solo_pos = get_solo_position(mock_map)

    # equivalent to to meter i.e. a lenght and finite
    assert solo_pos.data.xyz.unit.is_equivalent('m')
    assert all(np.isfinite(solo_pos.data.xyz))
    assert np.allclose(solo_pos.data.xyz, [-1.43160849e+07, -1.16297787e+08, 4.54713759e+07] * u.km)


# Testing to check if reprojection is correct by doing reprojecting map onto itself.
def test_map_reproject():
    map = Map(sample.AIA_094_IMAGE)
    observer = map.observer_coordinate
    reprojected_map = reproject_map(map, observer)
    assert np.allclose(map.observer_coordinate.lat, reprojected_map.observer_coordinate.lat)
    assert np.allclose(map.observer_coordinate.lon, reprojected_map.observer_coordinate.lon)
    assert np.allclose(map.observer_coordinate.radius, reprojected_map.observer_coordinate.radius)
    assert np.allclose(map.data, reprojected_map.data, atol=1e-7)



import os
import sys
import tarfile
from pathlib import Path

import stixpy


package_dir = Path(os.path.dirname(stixpy.__file__))
root_dir = package_dir.joinpath("data")


def _unzip_test_data():
    data_tar = root_dir / 'test_data.tar.gz'
    with tarfile.open(data_tar.as_posix(), 'r:gz') as tar:
        tar.extractall(root_dir.absolute())


_unzip_test_data()

_TEST_DATA = {
    'STIX_QL_BACKGROUND_TIMESERIES': 'solo_L1_stix-ql-background_20200506_V01.fits',
    'STIX_QL_CALIBRATION': 'solo_L1_stix-cal-energy_20200506_V01.fits',
    'STIX_QL_FLAREFLAG':  'solo_L1_stix-ql-ql-tmstatusflarelist_20200506_V01.fits',
    'STIX_QL_LIGHTCURVE_TIMESERIES': 'solo_L1_stix-ql-lightcurve_20200506_V01.fits',
    'STIX_QL_SPECTRA': 'solo_L1_stix-ql-spectra_20200506_V01.fits',
    'STIX_QL_VARIANCE_TIMESERIES': 'solo_L1_stix-ql-variance_20200506_V01.fits',
    'STIX_SCI_XRAY_SPEC':
        'solo_L1_stix-sci-xray-spec_20200505T235959-20200506T000019_V01_87031812-50886.fits',
    'STIX_SCI_XRAY_RPD':
        'solo_L1_stix-sci-xray-rpd_20200505T235959-20200506T000019_V01_87031808-50882.fits',
    'STIX_SCI_XRAY_CPD':
        'solo_L1_stix-sci-xray-cpd_20200505T235959-20200506T000019_V01_87031809-50883.fits',
    'STIX_SCI_XRAY_SCPD':
        'solo_L1_stix-sci-xray-scpd_20200505T235959-20200506T000019_V01_87031810-50884.fits',
    # 'STIX_SCI_XRAY_VIS':
    #     'solo_L1_stix-sci-xray-vis-87031811_20200505T235958-20200510T000014_V01_50885.fits'
}

__doc__ = ''
for k, v in _TEST_DATA.items():
    p = root_dir / v
    if not p.exists():
        raise ValueError(f'Test data {k}, {v} missing please try manually running _unzip_test_date()')
    setattr(sys.modules[__name__], k, str(p))
    __doc__ = __doc__ + f'   - ``{k}``\n'

__all__ = [*_TEST_DATA.keys()]



from sunpy.net import attrs as a
from sunpy.net.dataretriever import GenericClient
from sunpy.net.dataretriever.client import QueryResponse
from sunpy.time import TimeRange
from sunpy.util.scraper import Scraper

__all__ = ["STIXClient"]


class STIXClient(GenericClient):
    """
    A Fido client to search and download STIX data from the STIX instrument archive

    Examples
    --------
    >>> from sunpy.net import Fido, attrs as a
    >>> from stixpy.net.client import STIXClient
    >>> query = Fido.search(a.Time('2020-06-05', '2020-06-07'), a.Instrument.stix,
    ...                     a.stix.DataProduct.ql_lightcurve)  #doctest: +REMOTE_DATA
    >>> query  #doctest: +REMOTE_DATA
    <sunpy.net.fido_factory.UnifiedResponse object at ...>
    Results from 1 Provider:
    <BLANKLINE>
    3 Results from the STIXClient:
           Start Time               End Time        ...  DataProduct  Request ID
    ----------------------- ----------------------- ... ------------- ----------
    2020-06-05 00:00:00.000 2020-06-05 23:59:59.999 ... ql-lightcurve          -
    2020-06-06 00:00:00.000 2020-06-06 23:59:59.999 ... ql-lightcurve          -
    2020-06-07 00:00:00.000 2020-06-07 23:59:59.999 ... ql-lightcurve          -
    <BLANKLINE>
    <BLANKLINE>
    """
    baseurl = (r'http://pub099.cs.technik.fhnw.ch/data/fits/'
               r'{level}/{year:4d}/{month:02d}/{day:02d}/{datatype}/')
    ql_filename = r'solo_{level}_stix-{product}_\d{{8}}_V\d{{2}}.fits'
    sci_filename = (r'solo_{level}_stix-{product}_'
                    r'\d{{8}}T\d{{6}}-\d{{8}}T\d{{6}}_V\d{{2}}_\d+-\d{{5}}.fits')

    base_pattern = r'{}/{Level}/{year:4d}/{month:02d}/{day:02d}/{DataType}/'
    ql_pattern = r'solo_{Level}_{descriptor}_{time}_{ver}.fits'
    sci_pattern = r'solo_{Level}_{descriptor}_{start}-{end}_{Ver}_{Request}-{tc}.fits'

    required = {a.Time, a.Instrument}

    def search(self, *args, **kwargs):
        """
        Query this client for a list of results.

        Parameters
        ----------
        *args: `tuple`
         `sunpy.net.attrs` objects representing the query.
        **kwargs: `dict`
          Any extra keywords to refine the search.

        Returns
        -------
        A `QueryResponse` instance containing the query result.
        """
        matchdict = self._get_match_dict(*args, **kwargs)
        levels = matchdict['Level']

        metalist = []
        tr = TimeRange(matchdict['Start Time'], matchdict['End Time'])
        for date in tr.get_dates():
            year = date.datetime.year
            month = date.datetime.month
            day = date.datetime.day
            for level in levels:
                for datatype in matchdict['DataType']:
                    products = [p for p in matchdict['DataProduct']
                                if p.startswith(datatype.lower())]
                    for product in products:
                        if datatype.lower() == 'ql' and product.startswith('ql'):
                            url = self.baseurl + self.ql_filename
                            pattern = self.base_pattern + self.ql_pattern
                        elif datatype.lower() == 'sci' and product.startswith('sci'):
                            url = self.baseurl + self.sci_filename
                            pattern = self.base_pattern + self.sci_pattern
                        elif datatype.lower() == 'cal' and product.startswith('cal'):
                            url = self.baseurl + self.ql_filename
                            pattern = self.base_pattern + self.ql_pattern
                        elif datatype.lower() == 'aux':
                            url = self.baseurl + self.ql_filename
                            pattern = self.base_pattern + self.ql_pattern

                        url = url.format(level=level.upper(), year=year, month=month, day=day,
                                         datatype=datatype.upper(),
                                         product=product.replace('_', '-'))

                        scraper = Scraper(url, regex=True)
                        filesmeta = scraper._extract_files_meta(tr, extractor=pattern)
                        for i in filesmeta:
                            rowdict = self.post_search_hook(i, matchdict)
                            file_tr = rowdict.pop('tr', None)
                            if file_tr is not None:
                                if file_tr.start >= tr.start and file_tr.end <= tr.end:
                                    metalist.append(rowdict)
                            else:
                                metalist.append(rowdict)

        return QueryResponse(metalist, client=self)

    def post_search_hook(self, exdict, matchdict):
        rowdict = super().post_search_hook(exdict, matchdict)
        product = rowdict.pop('descriptor').strip('stix-')
        if rowdict.get('DataType') == 'QL':
            rowdict['DataProduct'] = product
            rowdict['Request ID'] = '-'
            rowdict.pop('ver')
            rowdict.pop('time')
        elif rowdict.get('DataType') == 'SCI':
            rowdict['DataProduct'] = product
            rowdict['Request ID'] = int(rowdict['Request'])
            ts = rowdict.pop('start')
            te = rowdict.pop('end')
            tr = TimeRange(ts, te)
            rowdict['tr'] = tr
            rowdict['Start Time'] = tr.start.iso
            rowdict['End Time'] = tr.end.iso
            rowdict.pop('tc')
            rowdict.pop('Request')
        return rowdict

    @classmethod
    def _attrs_module(cls):
        return 'stix', 'stixpy.net.attrs'

    @classmethod
    def register_values(cls):
        from sunpy.net import attrs
        adict = {attrs.Instrument: [('STIX', 'Spectrometer/Telescope for Imaging X-rays')],
                 attrs.Level: [('L0', 'STIX: Decommutated, uncompressed, uncalibrated data.'),
                               ('L1', 'STIX: Engineering and UTC time conversion .'),
                               ('L2', 'STIX: Calibrated data.')],
                 attrs.stix.DataType: [('QL', 'Quick Look'), ('SCI', 'Science Data'),
                                       ('CAL', 'Calibration'), ('AUX', 'Auxiliary')],
                 attrs.stix.DataProduct: [('cal_energy', 'Energy Calibration'),
                                          ('ql_lightcurve', 'Quick look light curve'),
                                          ('ql_background', 'Quick look background light curve'),
                                          ('ql_variance', 'Quick look variance curve'),
                                          ('ql_spectra', 'Quick look spectra'),
                                          ('ql_calibration_spectrum', 'Quick look energy '
                                                                      'calibration spectrum'),
                                          ('ql_flareflag',
                                           'Quick look energy calibration spectrum'),
                                          ('sci_xray_rpd', 'Raw Pixel Data'),
                                          ('sci_xray_cpd', 'Compressed Pixel Data'),
                                          ('sci_xray_scpd', 'Summed Compressed Pixel Data'),
                                          ('sci_xray_vis', 'Visibilities'),
                                          ('sci_xray_spec', 'Spectrogram'),
                                          ('aux_auxiliary', 'Auxilary ephermeris data')]
                 }
        return adict



from types import SimpleNamespace

import numpy as np

import astropy.units as u
from astropy.table import Table
from astropy.time import Time

from stixcore.calibration.energy import get_elut
from stixcore.calibration.grid import get_grid_internal_shadowing
from stixcore.calibration.livetime import get_livetime_fraction
from stixcore.config.reader import read_subc_params


def get_subcollimator_info():
    r"""
    Return resolutions, orientations and labels for sub-collimators.

    Returns
    -------

    """
    grids = {9: np.array([3, 20, 22]) - 1, 8: np.array([16, 14, 32]) - 1,
             7: np.array([21, 26, 4]) - 1, 6: np.array([24, 8, 28]) - 1,
             5: np.array([15, 27, 31]) - 1, 4: np.array([6, 30, 2]) - 1,
             3: np.array([25, 5, 23]) - 1, 2: np.array([7, 29, 1]) - 1,
             1: np.array([12, 19, 17]) - 1, 0: np.array([11, 13, 18]) - 1}

    labels = {i: [str(i+1)+'a', str(i)+'b', str(i)+'c'] for i in range(10)}

    res32 = np.zeros(32)
    res32[grids[9]] = 178.6
    res32[grids[8]] = 124.9
    res32[grids[7]] = 87.3
    res32[grids[6]] = 61.0
    res32[grids[5]] = 42.7
    res32[grids[4]] = 29.8
    res32[grids[3]] = 20.9
    res32[grids[2]] = 14.6
    res32[grids[1]] = 10.2
    res32[grids[0]] = 7.1
    res10 = [7.1, 10.2, 14.6, 20.9, 29.8, 42.7, 61.0, 87.3, 124.9, 178.6]
    o32 = np.zeros(32)
    o32[grids[9]] = [150, 90, 30]
    o32[grids[8]] = [170, 110, 50]
    o32[grids[7]] = [10, 130, 70]
    o32[grids[6]] = [30, 150, 90]
    o32[grids[5]] = [50, 170, 110]
    o32[grids[4]] = [70, 10, 130]
    o32[grids[3]] = [90, 30, 150]
    o32[grids[2]] = [110, 50, 170]
    o32[grids[1]] = [130, 70, 10]
    o32[grids[0]] = [150, 90, 30]

    # g03_10 = [g03, g04, g05, g06, g07, g08, g09, g10]
    # g01_10 = [g01, g02, g03, g04, g05, g06, g07, g08, g09, g10]
    # g_plot = [g10, g05, g09, g04, g08, g03, g07, g02, g06, g01]
    # l_plot = [l10, l05, l09, l04, l08, l03, l07, l02, l06, l01]

    res = SimpleNamespace(res32=res32, o32=o32, res10=res10, label=labels)

    return res


def create_visibility(pixel_data, time_range, energy_range, phase_center, no_shadowing=False):
    r"""
    Return visibility created from pixel data with in given time and energy range.


    Parameters
    ----------
    pixel_data
        Input pixel data
    time_range :
        Start and end times
    energy_range
        Start and end energies
    phase_center
        The coordinates of the phase center
    no_shadowing : bool optional
        If set to Ture turn grid shadowing correction off
    Returns
    -------
    `dict`
        Visibility data and sub-collimator information
    """
    t_ind = (pixel_data.times >= Time(time_range[0])) & (pixel_data.times <= Time(time_range[1]))
    e_ind = ((pixel_data.energies['e_low'] >= energy_range[0] * u.keV)
             & (pixel_data.energies['e_high'] <= energy_range[1] * u.keV))

    counts, counts_err, times, t_norm, energies = pixel_data.get_data(time_indices=t_ind,
                                                                      energy_indices=e_ind)

    # For the moment copied from idl
    trigger_to_detector = [0, 0, 7, 7, 2, 1, 1, 6, 6, 5, 2, 3, 3, 4, 4, 5, 13,
                           12, 12, 11, 11, 10, 13, 14, 14, 9, 9, 10, 15, 15, 8, 8]

    # Map the triggers to all 32 detectors
    triggers = pixel_data.data['triggers'][:, trigger_to_detector].astype(float)[...]

    # Update the triggers for the CFL and BK groups as the exposed area are very different
    # Sam / Ewan ??
    bkg_ratio = 1.1
    cfl_ratio = 1.33
    triggers[:, 8] = triggers[:, 8] / (1 + cfl_ratio) * cfl_ratio  # CFL
    triggers[:, 7] = triggers[:, 7] / (1 + cfl_ratio)  # Detector share with CFL

    triggers[:, 9] = triggers[:, 9] / (1 + bkg_ratio) * bkg_ratio  # BKG
    triggers[:, 15] = triggers[:, 15] / (1 + bkg_ratio)  # Detector shared with BKG

    _, livefrac, _ = get_livetime_fraction(
        triggers / pixel_data.data['timedel'].to('s').reshape(-1, 1))

    pixel_data.data['livefrac'] = livefrac

    elut = get_elut(pixel_data.time_range.center)
    real_dE = elut.e_width_actual[..., (pixel_data.energy_masks.masks[0] == 1)[1:-1]] * u.keV

    counts = pixel_data.data['counts']
    rate = (counts / (real_dE * u.keV * livefrac.reshape(45, 32, 1, 1)
                      * pixel_data.data['timedel'].to('s').reshape(-1, 1, 1, 1)))

    ct = counts[t_ind][..., e_ind]
    rate[t_ind][..., e_ind]

    # er = real_dE[..., e_ind].sum(-1)
    er = elut.e_width_actual[:, :, 3:7].sum(axis=-1) * u.keV
    lt = (livefrac * pixel_data.data['timedel'].reshape(-1, 1).to('s'))[t_ind].sum(axis=0)

    ct_sumed = ct.sum(axis=(0, 3)) / (er * lt.to('s').reshape(-1, 1))
    err_sumed = np.sqrt(ct.sum(axis=(0, 3))) / (er * lt.to('s').reshape(-1, 1))

    grid_shadowing = get_grid_internal_shadowing(phase_center)

    if not no_shadowing:
        ct_sumed = ct_sumed / grid_shadowing.reshape(-1, 1) / 4  # transmission grid ~ 0.5*0.5 = .25
        err_sumed = err_sumed / grid_shadowing.reshape(-1, 1) / 4

    abcd_rate = ct_sumed.reshape(-1, 3, 4)[:, [0, 1], :].sum(axis=1)

    # Taken from IDL
    pixel_areas = [0.096194997, 0.096194997, 0.096194997, 0.096194997, 0.096194997, 0.096194997,
                   0.096194997, 0.096194997, 0.010009999, 0.010009999, 0.010009999, 0.010009999]

    areas = np.tile(np.array(pixel_areas).reshape(1, -1), (32, 1)) * u.cm

    areas_summed = areas.reshape(-1, 3, 4)[:, [0, 1], :].sum(axis=1)

    aaaa = abcd_rate / areas_summed

    aaaa_err = np.sqrt(np.sum(err_sumed.reshape(-1, 3, 4)[:, [0, 1], :] ** 2, axis=1))
    aaaa_err = aaaa_err / areas_summed

    real = aaaa[:, 2] - aaaa[:, 0]
    imag = aaaa[:, 3] - aaaa[:, 1]

    real_err = np.sqrt(aaaa_err[:, 2] ** 2 + aaaa_err[:, 0] ** 2).value * real.unit
    imag_err = np.sqrt(aaaa_err[:, 3] ** 2 + aaaa_err[:, 1] ** 2).value * real.unit

    vis = get_visibility_info()
    vis['real'] = real
    vis['imag'] = imag
    vis['real_err'] = real_err
    vis['imag_err'] = imag_err
    vis['rate'] = aaaa
    vis['reate_err'] = aaaa_err

    return vis


def get_visibility_info(grid_separation=550 * u.mm):
    r"""
    Return the sub-collimator data and corresponding u, v points.

    Parameters
    ----------
    grid_separation : `float`, optional
        Front to read grid separation

    Returns
    -------

    """
    # imaging sub-collimators
    subc = read_subc_params()
    imaging_ind = np.where((subc['Grid Label'] != 'cfl') & (subc['Grid Label'] != 'bkg'))

    # filter out background monitor and flare locator
    subc_imaging = subc[imaging_ind]

    # take average of front and rear grid pitches (mm)
    pitch = (subc_imaging['Front Pitch'] + subc_imaging['Rear Pitch']) / 2.0

    # convert pitch from mm to arcsec
    # TODO check diff not using small angle approx
    pitch = (pitch / grid_separation).decompose() * u.rad

    # take average of front and rear grid orientation
    orientation = (subc_imaging['Front Orient'] + subc_imaging['Rear Orient']) / 2.0

    # calculate number of frequency components
    len(subc_imaging)

    # assign detector numbers to visibility index of subcollimator (isc)
    isc = subc_imaging['Det #']

    # assign the stix sc label for convenience
    label = subc_imaging['Grid Label']

    # save phase orientation of the grids to the visibility
    phase_sense = subc_imaging['Phase Sense']

    # calculate u and v
    uv = 1.0 / pitch.to('arcsec')
    uu = uv * np.cos(orientation.to('rad'))
    vv = uv * np.sin(orientation.to('rad'))

    # Add the livetime isc association. This gives the livetime pairings
    # vis.live_time = stx_ltpair_assignment(vis.isc)

    vis = {
        'pitch': pitch,
        'orientation': orientation,
        'isc': isc,
        'label': label,
        'phase_sense': phase_sense,
        'u': uu,
        'v': vv
    }
    return vis


def get_visibility_info_giordano():
    L1 = 550 * u.mm
    L2 = 47 * u.mm

    subc = read_subc_params()
    imaging_ind = np.arange(32)
    # np.where((subc['Grid Label'] != 'cfl') & (subc['Grid Label'] != 'bkg'))

    # filter out background monitor and flare locator
    subc_imaging = subc[imaging_ind]

    # take average of front and rear grid pitches (mm)
    # pitch = (subc_imaging['Front Pitch'] + subc_imaging['Rear Pitch']) / 2.0

    # convert pitch from mm to arcsec
    # TODO check diff not using small angle approx
    # pitch = (pitch / L1).decompose() * u.rad

    # take average of front and rear grid orientation
    # (subc_imaging['Front Orient'] + subc_imaging['Rear Orient']) / 2.0

    # calculate number of frequency components
    # len(subc_imaging)

    # assign detector numbers to visibility index of subcollimator (isc)
    # isc = subc_imaging['Det #']

    # assign the stix sc label for convenience
    subc_imaging['Grid Label']

    # save phase orientation of the grids to the visibility
    phase_sense = subc_imaging['Phase Sense']

    pitch_front = (1 / subc_imaging['Front Pitch'] * (L2 + L1)) * 1/u.rad
    pitch_rear = (1/subc_imaging['Rear Pitch'] * L2) * 1/u.rad

    uu = (np.cos(subc_imaging['Front Orient'].to('deg')) * pitch_front.to(1/u.arcsec)
          - np.cos(subc_imaging['Rear Orient'].to('deg')) * pitch_rear.to(1/u.arcsec))
    vv = (np.sin(subc_imaging['Front Orient'].to('deg')) * pitch_front.to(1/u.arcsec)
          - np.sin(subc_imaging['Rear Orient'].to('deg')) * pitch_rear.to(1/u.arcsec))

    uu = -uu * phase_sense
    vv = -vv * phase_sense

    return uu.to(1/u.arcsec), vv.to(1/u.arcsec)


def calibrate_visibility(vis, flare_location=(0, 0) * u.arcsec):
    """
    Return phase and amplitude calibrated visibilities.

    Parameters
    ----------
    vis :
        Uncalibrated Visibilities

    Returns
    -------

    """
    real_component = vis.pop('real')
    imag_component = vis.pop('imag')
    real_component_err = vis.pop('real_err')
    imag_component_err = vis.pop('imag_err')

    # Compute raw phases
    phase = np.arctan2(imag_component, real_component)

    # Compute amplitudes
    modulation_efficiency = np.pi ** 3 / (8 * np.sqrt(2))  # Modulation efficiency
    observed_amplitude = np.sqrt(real_component ** 2 + imag_component ** 2)
    calibrated_amplitude = observed_amplitude * modulation_efficiency

    # Compute error on amplitudes
    systematic_error = 5 * u.percent  # from G. Hurford 5% systematics
    amplitude_stat_error = ((np.sqrt((real_component / observed_amplitude * real_component_err) ** 2
                            + (imag_component / observed_amplitude * imag_component_err) ** 2))
                            * modulation_efficiency)
    amplitude_error = np.sqrt(amplitude_stat_error ** 2 +
                              (systematic_error * calibrated_amplitude)
                              .to(observed_amplitude.unit) ** 2)

    # Apply pixel correction
    phase += 46.1 * u.deg  # Center of large pixel in terms of phase of morie pattern
    # TODO add correction for small pixel

    # Grid correction factors
    phase_cal = Table.read('/usr/local/ssw/so/stix/dbase/demo/vis_demo/GridCorrection.csv',
                           header_start=2, data_start=3)
    grid_corr = phase_cal['Phase correction factor'] * u.deg

    # Apply grid correction
    phase += grid_corr

    # Phase correction
    phase_cal = Table.read('/usr/local/ssw/so/stix/dbase/demo/vis_demo/PhaseCorrFactors.csv',
                           header_start=3, data_start=4)
    phase_corr = phase_cal['Phase correction factor'] * u.deg

    # Apply grid correction
    phase += phase_corr

    # Compute error on phases
    phase_error = amplitude_error / calibrated_amplitude

    imaging_detector_indices = vis['isc'] - 1

    vis = SimpleNamespace(amplitude=calibrated_amplitude[imaging_detector_indices],
                          amplitude_error=amplitude_error[imaging_detector_indices],
                          phase=phase[imaging_detector_indices],
                          phase_error=phase_error[imaging_detector_indices],
                          **vis)

    return vis


def correct_phase_projection(vis, flare_xy):
    r"""
    Correct visibilities for projection of the grids onto the detectors.

    The rear grids are not in direct contact with the detectors so the phase of the visibilities
    needs to be project onto the grid and this is dependant on the position of the xray emission

    Parameters
    ----------
    vis
        Input visibilities
    flare_xy
        Position of the flare

    Returns
    -------

    """
    # Phase projection correction
    l1 = 550 * u.mm  # seperation of front rear grid
    l2 = 47 * u.mm  # seperation of rear grid and detector
    # TODO check how this works out to degress
    vis.phase -= (flare_xy[1].to_value('arcsec') * 360. * np.pi
                  / (180. * 3600. * 8.8 * u.mm) * (l2 + l1 / 2.0)) * u.deg

    vis.u *= -vis.phase_sense
    vis.v *= -vis.phase_sense

    # Compute real and imaginary part of the visibility
    vis.obsvis = vis.amplitude * (np.cos(vis.phase.to('rad')) + np.sin(vis.phase.to('rad')) * 1j)

    # Add phase factor for shifting the flare_xy
    map_center = flare_xy - [26.1, 58.2] * u.arcsec
    # Subtract Frederic's mean shift values

    # TODO check the u, v why is flare_xy[1] used with u (prob stix vs spacecraft orientation
    phase_mapcenter = -2 * np.pi * (map_center[1] * vis.u - map_center[0] * vis.v) * u.rad
    vis.obsvis *= (np.cos(phase_mapcenter) + np.sin(phase_mapcenter) * 1j)

    return vis


def sas_map_center():
    # receter map at 0,0 taking account of mean or actual sas sol
    pass



import astropy.units as u
import numpy as np
from types import SimpleNamespace

from sunpy.io.special import read_genx
from stixcore.calibration.transmission import Transmission


def get_srm():
    r"""
    Return the spectromber response matrix (SRM) by combing the attenuation with the detetor respoonse matrix (DRM)

    Returns
    -------

    """
    drm_save = read_genx('/Users/shane/Projects/STIX/git/stix_drm_20220713.genx')
    drm = drm_save['SAVEGEN0']['SMATRIX'] * u.count/u.keV/u.photon
    energies_in = drm_save['SAVEGEN0']['EDGES_IN'] * u.keV
    energies_in_width = np.diff(energies_in)
    energies_in_mean = energies_in[:-1] + energies_in_width/2
    trans = Transmission()
    tot_trans = trans.get_transmission(energies=energies_in_mean)
    energies_out = drm_save['SAVEGEN0']['EDGES_OUT'] * u.keV
    energies_out_width = drm_save['SAVEGEN0']['EWIDTH'] * u.keV
    energies_out_mean = drm_save['SAVEGEN0']['EMEAN'] * u.keV
    attenuation = tot_trans['det-0']
    srm = (attenuation.reshape(-1, 1) * drm * energies_out_width) / 4  # avg grid transmission

    res = SimpleNamespace(drm=drm, srm=srm, attenuation=attenuation, energies_in=energies_in,
                          energies_in_width=energies_in_width, energies_in_mean=energies_in_mean,
                          energies_out=energies_out, energies_out_width=energies_out_width,
                          energies_out_mean=energies_out_mean, area=1*u.cm)
    return res


def get_pixel_srm():
    pass



import os
import pathlib
from collections import OrderedDict
from urllib.request import Request

import numpy as np

import astropy.io.fits
from astropy.wcs import WCS

from sunpy import log
from sunpy.data import cache
from sunpy.io.file_tools import read_file
from sunpy.io.header import FileHeader
from sunpy.map.compositemap import CompositeMap
from sunpy.map.mapbase import GenericMap, MapMetaValidationError
from sunpy.map.mapsequence import MapSequence
from sunpy.util import expand_list
from sunpy.util.datatype_factory_base import (
    BasicRegistrationFactory,
    MultipleMatchError,
    NoMatchError,
    ValidationFunctionError,
)
from sunpy.util.exceptions import NoMapsInFileError, warn_user
from sunpy.util.functools import seconddispatch
from sunpy.util.io import is_url, parse_path, possibly_a_path
from sunpy.util.metadata import MetaDict
from sunpy.util.types import DatabaseEntryType

SUPPORTED_ARRAY_TYPES = (np.ndarray,)
try:
    import dask.array
    SUPPORTED_ARRAY_TYPES += (dask.array.Array,)
except ImportError:
    pass

__all__ = ['Product', 'ProductFactory']


class ProductFactory(BasicRegistrationFactory):
    """
    A factory for generating stix data products

    Parameters
    ----------
    \\*inputs
        Inputs to parse for map objects. See the examples section for a
        detailed list of accepted inputs.
    sequence : `bool`, optional
        Return a `sunpy.map.MapSequence` object comprised of all the parsed maps.
    composite : `bool`, optional
        Return a `sunpy.map.CompositeMap` object comprised of all the parsed maps.
    Returns
    -------
    `sunpy.map.GenericMap`
        If the input results in a singular map object, then that is returned.
    `list` of `~sunpy.map.GenericMap`
        If multiple inputs are given and ``sequence=False`` and ``composite=False``
        (the default) then a list of `~sunpy.map.GenericMap` objects will be
        returned.
    `sunpy.map.MapSequence`
        If the input corresponds to multiple maps and ``sequence=True`` is set,
        then a `~sunpy.map.MapSequence` object is returned.
    `sunpy.map.CompositeMap`
        If the input corresponds to multiple maps and ``composite=True`` is set,
        then a `~sunpy.map.CompositeMap` object is returned.
    Examples
    --------
    >>> import sunpy.map
    >>> from astropy.io import fits
    >>> import sunpy.data.sample  # doctest: +REMOTE_DATA
    >>> mymap = sunpy.map.Map(sunpy.data.sample.AIA_171_IMAGE)  # doctest: +REMOTE_DATA
    The SunPy Map factory accepts a wide variety of inputs for creating maps
    * Preloaded tuples of (data, header) pairs
    >>> mymap = sunpy.map.Map((data, header))   # doctest: +SKIP
    headers are some base of `dict` or `collections.OrderedDict`, including
    `sunpy.io.header.FileHeader` or `sunpy.util.metadata.MetaDict` classes.
    * data, header pairs, not in tuples
    >>> mymap = sunpy.map.Map(data, header)   # doctest: +SKIP
    * data, wcs object, in tuple
    >>> from astropy.wcs import WCS
    >>> wcs = WCS(sunpy.data.sample.AIA_171_ROLL_IMAGE)  # doctest: +SKIP
    >>> data = fits.getdata(sunpy.data.sample.AIA_171_ROLL_IMAGE)  # doctest: +SKIP
    >>> mymap = sunpy.map.Map((data, wcs))  # doctest: +SKIP
    * data, wcs object, not in tuple
    >>> from astropy.wcs import WCS
    >>> wcs = WCS(sunpy.data.sample.AIA_171_ROLL_IMAGE)  # doctest: +SKIP
    >>> data = fits.getdata(sunpy.data.sample.AIA_171_ROLL_IMAGE)  # doctest: +SKIP
    >>> mymap = sunpy.map.Map(data, wcs)  # doctest: +SKIP
    * File names
    >>> mymap = sunpy.map.Map('file1.fits')   # doctest: +SKIP
    * All fits files in a directory by giving a directory
    >>> mymap = sunpy.map.Map('local_dir/sub_dir')   # doctest: +SKIP
    * A filesystem path expressed as a `pathlib.Path`
    >>> import pathlib
    >>> mymap = sunpy.map.Map(pathlib.Path('file1.fits'))   # doctest: +SKIP
    >>> sub_dir = pathlib.Path('local_dir/sub_dir')
    >>> mymap = sunpy.map.Map(sub_dir)   # doctest: +SKIP
    >>> mymap = sunpy.map.Map(sub_dir / 'file3.fits')   # doctest: +SKIP
    * Some regex globs
    >>> mymap = sunpy.map.Map('eit_*.fits')   # doctest: +SKIP
    * URLs
    >>> mymap = sunpy.map.Map(url_str)   # doctest: +SKIP
    * DatabaseEntry
    >>> mymap = sunpy.map.Map(db_result)   # doctest: +SKIP
    * Lists of any of the above
    >>> mymap = sunpy.map.Map(['file1.fits', 'file2.fits', 'file3.fits', 'directory1/'])  # doctest: +SKIP
    * Any mixture of the above not in a list
    >>> mymap = sunpy.map.Map(((data, header), data2, header2, 'file1.fits', url_str, 'eit_*.fits'))  # doctest: +SKIP
    """

    def _read_file(self, fname, **kwargs):
        """
        Read in a file name and return the list of (data, meta) pairs in that file.
        """
        # File gets read here. This needs to be generic enough to seamlessly
        # call a fits file or a jpeg2k file, etc
        # NOTE: use os.fspath so that fname can be either a str or pathlib.Path
        # This can be removed once read_file supports pathlib.Path
        log.debug(f'Reading {fname}')
        try:
            pairs = read_file(os.fspath(fname), **kwargs)
        except Exception as e:
            msg = f"Failed to read {fname}."
            raise IOError(msg) from e

        new_pairs = []
        for pair in pairs:
            filedata, filemeta = pair
            assert isinstance(filemeta, FileHeader)
            # This tests that the data is more than 1D
            if len(np.shape(filedata)) > 1:
                data = filedata
                meta = MetaDict(filemeta)
                new_pairs.append((data, meta))

        if not new_pairs:
            raise NoMapsInFileError(f"Found no HDUs with >= 2D data in '{fname}'.")

        return new_pairs

    def _validate_meta(self, meta):
        """
        Validate a meta argument.
        """
        if isinstance(meta, astropy.io.fits.header.Header):
            return True
        elif isinstance(meta, dict):
            return True
        else:
            return False

    def _parse_args(self, *args, silence_errors=False, **kwargs):
        """
        Parses an args list into data-header pairs.
        args can contain any mixture of the following entries:
        * tuples of data,header
        * data, header not in a tuple
        * data, wcs object in a tuple
        * data, wcs object not in a tuple
        * filename, as a str or pathlib.Path, which will be read
        * directory, as a str or pathlib.Path, from which all files will be read
        * glob, from which all files will be read
        * url, which will be downloaded and read
        * lists containing any of the above.
        Examples
        --------
        self._parse_args(data, header,
                         (data, header),
                         ['file1', 'file2', 'file3'],
                         'file4',
                         'directory1',
                         '*.fits')
        """
        # Account for nested lists of items
        args = expand_list(args)

        # Sanitise the input so that each 'type' of input corresponds to a different
        # class, so single dispatch can be used later
        nargs = len(args)
        i = 0
        while i < nargs:
            arg = args[i]
            if isinstance(arg, SUPPORTED_ARRAY_TYPES):
                # The next two items are data and a header
                data = args.pop(i)
                header = args.pop(i)
                args.insert(i, (data, header))
                nargs -= 1
            elif isinstance(arg, str) and is_url(arg):
                # Repalce URL string with a Request object to dispatch on later
                args[i] = Request(arg)
            elif possibly_a_path(arg):
                # Repalce path strings with Path objects
                args[i] = pathlib.Path(arg)
            i += 1

        # Parse the arguments
        # Note that this list can also contain GenericMaps if they are directly given to the factory
        data_header_pairs = []
        for arg in args:
            try:
                data_header_pairs += self._parse_arg(arg, **kwargs)
            except NoMapsInFileError as e:
                if not silence_errors:
                    raise
                warn_user(f"One of the arguments failed to parse with error: {e}")

        return data_header_pairs

    # Note that post python 3.8 this can be @functools.singledispatchmethod
    @seconddispatch
    def _parse_arg(self, arg, **kwargs):
        """
        Take a factory input and parse into (data, header) pairs.
        Must return a list, even if only one pair is returned.
        """
        raise ValueError(f"Invalid input: {arg}")

    @_parse_arg.register(tuple)
    def _parse_tuple(self, arg, **kwargs):
        # Data-header or data-WCS pair
        data, header = arg
        if isinstance(header, WCS):
            header = header.to_header()

        pair = data, header
        if self._validate_meta(header):
            pair = (data, OrderedDict(header))
        return [pair]

    @_parse_arg.register(DatabaseEntryType)
    def _parse_dbase(self, arg, **kwargs):
        return self._read_file(arg.path, **kwargs)

    @_parse_arg.register(GenericMap)
    def _parse_map(self, arg, **kwargs):
        return [arg]

    @_parse_arg.register(Request)
    def _parse_url(self, arg, **kwargs):
        url = arg.full_url
        path = str(cache.download(url).absolute())
        pairs = self._read_file(path, **kwargs)
        return pairs

    @_parse_arg.register(pathlib.Path)
    def _parse_path(self, arg, **kwargs):
        return parse_path(arg, self._read_file, **kwargs)

    def __call__(self, *args, composite=False, sequence=False, silence_errors=False, **kwargs):
        """ Method for running the factory. Takes arbitrary arguments and
        keyword arguments and passes them to a sequence of pre-registered types
        to determine which is the correct Map-type to build.
        Arguments args and kwargs are passed through to the validation
        function and to the constructor for the final type. For Map types,
        validation function must take a data-header pair as an argument.
        Parameters
        ----------
        composite : `bool`, optional
            Indicates if collection of maps should be returned as a `~sunpy.map.CompositeMap`.
            Default is `False`.
        sequence : `bool`, optional
            Indicates if collection of maps should be returned as a `sunpy.map.MapSequence`.
            Default is `False`.
        silence_errors : `bool`, optional
            If set, ignore data-header pairs which cause an exception.
            Default is ``False``.
        Notes
        -----
        Extra keyword arguments are passed through to `sunpy.io.read_file` such
        as `memmap` for FITS files.
        """
        data_header_pairs = self._parse_args(*args, silence_errors=silence_errors, **kwargs)
        new_maps = list()

        # Loop over each registered type and check to see if WidgetType
        # matches the arguments.  If it does, use that type.
        for pair in data_header_pairs:
            if isinstance(pair, GenericMap):
                new_maps.append(pair)
                continue
            data, header = pair
            meta = MetaDict(header)

            try:
                new_map = self._check_registered_widgets(data, meta, **kwargs)
                new_maps.append(new_map)
            except (NoMatchError, MultipleMatchError,
                    ValidationFunctionError, MapMetaValidationError) as e:
                if not silence_errors:
                    raise
                warn_user(f"One of the data, header pairs failed to validate with: {e}")

        if not len(new_maps):
            raise RuntimeError('No maps loaded')

        # If the list is meant to be a sequence, instantiate a map sequence
        if sequence:
            return MapSequence(new_maps, **kwargs)

        # If the list is meant to be a composite map, instantiate one
        if composite:
            return CompositeMap(new_maps, **kwargs)

        if len(new_maps) == 1:
            return new_maps[0]

        return new_maps

    def _check_registered_widgets(self, data, meta, **kwargs):

        candidate_widget_types = list()

        for key in self.registry:

            # Call the registered validation function for each registered class
            if self.registry[key](data, meta, **kwargs):
                candidate_widget_types.append(key)

        n_matches = len(candidate_widget_types)

        if n_matches == 0:
            if self.default_widget_type is None:
                raise NoMatchError("No types match specified arguments and no default is set.")
            else:
                candidate_widget_types = [self.default_widget_type]
        elif n_matches > 1:
            raise MultipleMatchError("Too many candidate types identified "
                                     f"({candidate_widget_types}). "
                                     "Specify enough keywords to guarantee unique type "
                                     "identification.")

        # Only one is found
        WidgetType = candidate_widget_types[0]

        return WidgetType(data, meta, **kwargs)


class InvalidMapInput(ValueError):
    """Exception to raise when input variable is not a Map instance and does
    not point to a valid Map input file."""


class InvalidMapType(ValueError):
    """Exception to raise when an invalid type of map is requested with Map
    """


class NoMapsFound(ValueError):
    """Exception to raise when input does not point to any valid maps or files
    """


Map = MapFactory(registry=GenericMap._registry, default_widget_type=GenericMap,
                 additional_validation_functions=['is_datasource_for'])


import numpy as np

from stixcore.calibration.livetime import get_livetime_fraction


def test_get_livetime():
    eta = 2.5e-6
    tau = 12.5e-6
    ph_in = np.arange(1000) * 1e3

    # Simulate normal trigger process
    trig1 = ph_in / (1 + ph_in * (tau + eta))
    # Two photon contribution
    trig2 = trig1 * np.exp(-1 * ph_in * eta)

    livetime_fraction1, livetime_fraction2, ph_out = get_livetime_fraction(trig1, eta=eta, tau=tau)

    assert np.allclose(ph_in, ph_out)
    assert np.allclose(ph_in, trig1/livetime_fraction1)
    assert np.allclose(ph_in, trig2/livetime_fraction2)




import numpy as np
from numpy.testing import assert_allclose

import astropy.units as u

from stixcore.calibration.transmission import Transmission


def test_transmission_create_material():
    hydrogen = Transmission.create_material(name='hydrogen', fractional_masses={'H': 1.0},
                                            thickness=1*u.cm, density=8.375e-05*u.g/u.cm**3)

    assert hydrogen.name == 'hydrogen'
    assert hydrogen.thickness == 1.0 * u.cm
    assert hydrogen.density == 8.375e-05*u.g/u.cm**3
    # taken from https://physics.nist.gov/PhysRefData/XrayMassCoef/ElemTab/z01.html
    ref_attenuation_coefficient = [7.217E+00, 2.944E-01, 3.254E-02] * u.cm**2/u.g
    attenuation_coefficient = hydrogen.mass_attenuation_coefficient.func([1e-3, 1e-1, 1e+1] * u.MeV)
    assert np.allclose(attenuation_coefficient, ref_attenuation_coefficient)
    transmission = np.exp(-hydrogen.density*hydrogen.thickness*attenuation_coefficient)
    assert np.allclose(transmission, hydrogen.transmission([1e-3, 1e-1, 1e+1] * u.MeV))


def test_transmission_get_transmission():
    # test_materials = OrderedDict([('test1', [('al', 1 * u.mm), ('be', 2*u.mm)]),
    #                           ('test2', [('be', 1*u.mm)])])
    #
    # test_componenets = OrderedDict([('al', ({'Al': 1.0}, 2.7 * u.g/u.cm**3)),
    #                           ('be', ({'Be': 1.0}, 1.85 * u.g/u.cm**3))])

    trans = Transmission()
    res = trans.get_transmission()
    assert len(res.columns) == 34
    assert len(res) == 31
    assert res['energies'][0] == 4.0
    assert res['energies'][-1] == 150.0
    assert_allclose(res['det-0'][0], 1.2304064853433878e-05)
    assert_allclose(res['det-0'][-1], 0.9223594048282234)
    assert_allclose(res['det-10'][0], 2.866929141727759e-06)
    assert_allclose(res['det-10'][-1], 0.9186555303212736)



"""
One overriding requirement for STIX is that the incoming photon rate can be determined with high
accuracy from the measured count rate. As such the dead time or equivalently the live time must be
accurately determined from the trigger rates. The live time fraction is defined as the ratio of the
event trigger rate to the photon rate :math:`R_{trigger}/R_{photon}`. The dead time is primarily
controlled by two timescales the readout time :math:`\\tau` and latency :math:`\\eta`.

These timescales correspond to two distinct contributions to the effective live time of a group of
ASICs. First, photons arriving during the read-out sequence of a previously triggered ASIC of that
group will not be registered as the trigger circuit is inactive during the readout. Second, of those
events that did trigger, a certain fraction will suffer from multiple hits and will be discarded,
if one or more additional triggers occur during the latency time before the read-out. The live time
fraction the ratio of the event rate to the photon rate is thus given by

.. math::

    R_{photon} = \\frac{R_{trigger}}{1 -  R_{trigger}(\\tau+\\eta)}

    \\text{livetime fraction} = \\frac{e^{-\\eta R_{photon}}}{ 1 + R_{photon}(\\tau+\\eta)}

where :math:`R_{trigger}` is the observed trigger rate and :math:`R_{photon}` is the derived
incident photon rate.

The plot below shows the behaviour of the trigger rate taking these effects into account.

.. plot::

    import numpy as np

    from matplotlib import pyplot as plt

    from stixcore.calibration.livetime import get_livetime_fraction

    eta = 2.5e-6
    tau = 12.5e-6
    ph_in = np.logspace(3, 6, 50)
    trig1 = ph_in / (1 + ph_in*(tau+eta) )
    trig2 = trig1*np.exp(-1*ph_in*eta)

    livetime_fraction1, livetime_fraction2, ph_out = get_livetime_fraction(trig1, eta=eta, tau=tau)

    plt.plot(ph_in, ph_in,  label='Photons In')
    plt.loglog()
    plt.xlim(1e3, 1e6)
    plt.ylim(1e3, 1e6)
    plt.plot(ph_in, trig1, '-.', label='Triggers')
    plt.plot(ph_in, trig2, ':',  label='Triggers (inc two photon)')
    plt.plot(ph_in, trig2/livetime_fraction2, '+',
             label='Photons derived from Triggers (inc two photon)')
    plt.xlabel('Incoming Photon Rate')
    plt.ylabel('Trigger Rate')
    plt.title(rf'$\\tau =$ {tau} s, $\\eta =$ {eta} s')
    plt.legend()
    plt.show()

References
----------

STIX-TN-0015-ETH_I1R0_Caliste_Rates
"""
import numpy as np

import astropy.units as u

__all__ = ['get_livetime_fraction']


def get_livetime_fraction(trigger_rate, *, eta=3.91*u.us, tau=11*u.us):
    """
    Return the live time fraction for the given trigger rate.

    Parameters
    ----------
    trigger_rate : `float`
        Event trigger rate
    eta : `float`
        Latency per event
    tau : `float`
        Readout time per event

    Returns
    -------
    `float`, `float`, `float`:
        The live time fraction
    """
    photons_in = trigger_rate / (1. - trigger_rate * (tau + eta))
    livetime_fraction = 1 / (1. + (tau + eta) * photons_in)
    two_photon = np.exp(-eta * photons_in) * livetime_fraction
    return livetime_fraction, two_photon, photons_in



"""
Grid Calibration
"""
from pathlib import Path

import numpy as np

import astropy.units as u
from astropy.table import Table


def get_grid_internal_shadowing(xy_flare):
    r"""
    Return the grid transmission for the 32 sub-collimators



    Read the grid tables and calculates grid transmission for 32 sub-collimators. For finest grids
    only transmission without correction for internal shadowing

    Parameters
    ----------
    xy_flare

    Returns
    -------

    """
    column_names = ['sc', 'p', 'o', 'phase', 'slit', 'grad', 'rms', 'thick', 'bwidth', 'bpitch']

    root = Path(__file__).parent.parent
    grid_info = Path(root, *['config', 'data', 'common', 'grid'])
    front = Table.read(grid_info / 'grid_param_front.txt', format='ascii', names=column_names)
    rear = Table.read(grid_info / 'grid_param_rear.txt', format='ascii', names=column_names)

    xy_flare_stix = np.array([xy_flare[1].value, -xy_flare[0].value])*u.arcsec
    # print,xy_flare
    # print,xy_flare_stix

    # radial
    r_flare = np.sqrt((xy_flare_stix**2).sum())
    # angle from +y direction (definition used by Matej in grid table)
    # o_flare=abs(asin(xy_flare_stix[1]/r_flare)/!pi*180)
    # o_flare=90-abs(asin(xy_flare_stix[1]/r_flare)/!pi*180)
    # new October 26

    # Paolo Massa (September 2021): corrected the definition of 'o_flare'

    if (xy_flare_stix[0] >= 0) and (xy_flare_stix[1] >= 0):
        o_flare = np.arccos(xy_flare_stix[1]/r_flare)
    if (xy_flare_stix[0] >= 0) and (xy_flare_stix[1] <= 0):
        o_flare = 90*u.deg+np.arcsin(np.abs(xy_flare_stix[1])/r_flare)
    if (xy_flare_stix[0] <= 0) and (xy_flare_stix[1] <= 0):
        o_flare = np.arccos(np.abs(xy_flare_stix[1])/r_flare)
    if (xy_flare_stix[0] <= 0) and (xy_flare_stix[1] >= 0):
        o_flare = 90*u.deg + np.arcsin(xy_flare_stix[1]/r_flare)

    # if (xy_flare_stix[0] >= 0) and (xy_flare_stix[1] >= 0):
    #     o_flare = np.arccos(xy_flare_stix[1] / r_flare)
    # if (xy_flare_stix[0] >= 0) and (xy_flare_stix[1] <= 0):
    #     o_flare = np.pi*u.rad/2 - np.arcsin(abs(xy_flare_stix[1]) / r_flare)
    # if (xy_flare_stix[0] <= 0) and (xy_flare_stix[1] <= 0):
    #     o_flare = -np.pi*u.rad/2 + np.arcsin(xy_flare_stix[1] / r_flare)
    # if (xy_flare_stix[0] <= 0) and (xy_flare_stix[1] >= 0):
    #     o_flare = -np.arccos(xy_flare_stix[1] / r_flare)

    # Paolo Massa (September 2021): corrected the definition of 'rel_f' and 'rel_r'. Indeed, the
    # grids in this case are considered looking towards the Sun. A minus sign is added to
    # 'front['o']' and 'rear['o']' to keep into account the different definition of the orientation
    # angle used in Matej's grid table

    #  #orientations of grids relative to flare
    #  rel_f=abs(front['o']-o_flare)
    #  rel_r=abs(rear['o']-o_flare)
    rel_f = np.abs((180 - front['o'])*u.deg - o_flare)
    rel_r = np.abs((180 - rear['o'])*u.deg - o_flare)

    # correction only applies to component normal to the grid orientation
    cor_f = np.sin(rel_f)
    cor_r = np.sin(rel_r)

    # nominal thickness with glue
    nom_h = 0.42  # * u.mm
    # maximal internal grid shadowing when see for grids with orientation of 90 degrees from
    # center to flare line
    max_c = np.tan(r_flare)*nom_h
    # the shadow in mm for each of the entries# this value will be subtracted from the nominal
    # slit width (see below)
    shadow_r = cor_r * max_c
    shadow_f = cor_f * max_c

    # for bridges correction is largest for grids with orientation parallel to the center-flare
    # line
    np.cos(rel_f)
    np.cos(rel_r)
    # the shadow to be added to bridge width (see below)
    cor_r * max_c
    cor_f * max_c

    gtrans32f = np.zeros(32) - 2
    gtrans32r = np.zeros(32) - 2
    btrans32f = np.zeros(32) + 1
    btrans32r = np.zeros(32) + 1

    # for the case the flare location is at [0,0], the code above gives NAN, but correction for
    # internal shadowing should be zero
    if xy_flare[0].to_value('arcsec') == 0 and xy_flare[1].to_value('arcsec') == 0:
        shadow_f = np.zeros(39)
        shadow_r = np.zeros(39)

    for i in range(32):
        this_list = np.where(rear['sc'] == i + 1)[0]
        # print,i,front['sc'][this_list]
        # single layer: transmission = slit/pitch
        if (this_list.size == 1) and (this_list[0] != -1):
            # gtrans32f[i]=front['slit'](this_list[0])/front['p'](this_list[0])
            gtrans32f[i] = ((front['slit'][this_list[0]] - shadow_f[this_list[0]])
                            / front['p'][this_list[0]])
            # gtrans32r[i]=rear['slit'](this_list[0])/rear['p'](this_list[0])
            gtrans32r[i] = ((rear['slit'][this_list[0]] - shadow_r[this_list[0]])
                            / rear['p'][this_list[0]])
            if front['bpitch'][this_list[0]] != 0.0:
                btrans32f[i] = 1 - front['bwidth'][this_list[0]] / front['bpitch'][this_list[0]]
                btrans32r[i] = 1 - rear['bwidth'][this_list[0]] / rear['bpitch'][this_list[0]]

        # multi layer grids:
        if this_list.size >= 2:
            # slats in each layer (covered part)
            this_slat_f_each = front['p'][this_list] - front['slit'][this_list]
            # sum of all slats gives lengths of total covered part
            this_slat_f_total = this_slat_f_each.sum()
            # transmission is 1 minus covered/pitch
            gtrans32f[i] = 1. - this_slat_f_total / np.average(front['p'][this_list])
            # slats in each layer (covered part)
            this_slat_r_each = rear['p'][this_list] - rear['slit'][this_list]
            # sum of all slats gives lengths of total covered part
            this_slat_r_total = this_slat_r_each.sum()
            # transmission is 1 minus covered/pitch
            gtrans32r[i] = 1. - this_slat_r_total / np.average(rear['p'][this_list])
            # bridge: only one value in the table
            btrans32f[i] = 1 - front['bwidth'][this_list[0]] / front['bpitch'][this_list[0]]
            btrans32r[i] = 1 - rear['bwidth'][this_list[0]] / rear['bpitch'][this_list[0]]

    # no bridge measurement for 5 front use 0.05, the nominal value
    # (in file it is set to 0.5 to mark the missing value)
    btrans32f[4] = 1. - 0.05 / front['bpitch'][4]
    # could apply averaged deviation from nominal instead?

    gtrans32 = gtrans32f * gtrans32r * btrans32f * btrans32r
    # gtrans32_no_b = gtrans32f * gtrans32r

    # set CFL and BKG to zero for now
    gtrans32[8:10] = 0
    return gtrans32



import re
import traceback
import warnings

import astropy.coordinates as coord
import astropy.units as u
import numpy as np

from astropy.coordinates import frame_transform_graph
from astropy.coordinates.baseframe import BaseCoordinateFrame
from astropy.coordinates.errors import ConvertError
from astropy.coordinates.matrix_utilities import rotation_matrix, matrix_product, matrix_transpose
from astropy.coordinates.representation import UnitSphericalRepresentation, SphericalRepresentation
from astropy.io import fits
from astropy.table import QTable
from astropy.time import Time
from sunpy.coordinates.frames import HeliographicStonyhurst, Helioprojective, \
    SunPyBaseCoordinateFrame
from sunpy.coordinates.transformations import _check_observer_defined
from sunpy.net import Fido, attrs as a
from sunpy.util.exceptions import warn_user

from stixcore.util.logging import get_logger
from stixpy.net.client import STIXClient  # noqa

logger = get_logger(__name__, 'DEBUG')

__all__ = ['STIXImagingFrame', '_get_rotation_matrix_and_position', 'hpc_to_stixim', 'stixim_to_hpc']

STIX_X_SHIFT = 26.1 * u.arcsec  # fall back to this when non sas solution available
STIX_Y_SHIFT = 58.2 * u.arcsec  # fall back to this when non sas solution available
STIX_X_OFFSET = 60.0 * u.arcsec  # remaining offset after SAS solution
STIX_Y_OFFSET = 8.0 * u.arcsec  # remaining offset after SAS solution


class STIXImagingFrame(SunPyBaseCoordinateFrame):
    r"""
    STIX Imaging Frame

    - ``x`` (aka "theta_x") along the pixel rows (e.g. 0, 1, 2, 3; 4, 5, 6, 8).
    - ``y`` (aka "theta_y") along the pixel columns (e.g. A, B, C, D).
    - ``distance`` is the Sun-observer distance.

    Aligned with SIIX 'pixels' +X corresponds direction along pixel row toward pixels
    0, 4 and +Y corresponds direction along colums towards pixels 0, 1, 2, 3.

    .. code-block:: text

      Col\Row
            ---------
           | 3  | 7  |
        D  |   _|_   |
           |  |11 |  |
            ---------
           | 2  | 6  |
        C  |   _|_   |
           |  |10 |  |
            ---------
           | 1  | 5  |
        B  |   _|_   |
           |  | 9 |  |
            ---------
           | 0  | 4  |
        A  |   _|_   |
           |  | 8 |  |
            ---------

    """

    frame_specific_representation_info = {
        coord.SphericalRepresentation: [
            coord.RepresentationMapping('lon', 'x', u.arcsec),
            coord.RepresentationMapping('lat', 'y', u.arcsec),
            coord.RepresentationMapping('distance', 'distance')]

    }

    # def make_3d(self):
    #     """
    #     This method calculates the third coordinate of the Helioprojective
    #     frame. It assumes that the coordinate point is on the surface of the Sun.
    #     If a point in the frame is off limb then NaN will be returned.
    #     Returns
    #     -------
    #     new_frame : `~sunpy.coordinates.frames.Helioprojective`
    #         A new frame instance with all the attributes of the original but
    #         now with a third coordinate.
    #     """
    #     # Skip if we already are 3D
    #     if not self._is_2d:
    #         return self
    #
    #     if not isinstance(self.observer, BaseCoordinateFrame):
    #         raise ConvertError("Cannot calculate distance to the Sun "
    #                            f"for observer '{self.observer}' "
    #                            "without `obstime` being specified.")
    #
    #     rep = self.represent_as(UnitSphericalRepresentation)
    #     lat, lon = rep.lat, rep.lon
    #
    #     # Check for the use of floats with lower precision than the native Python float
    #     if not set([lon.dtype.type, lat.dtype.type]).issubset([float, np.float64, np.longdouble]):
    #         warn_user("The Helioprojective component values appear to be lower "
    #                   "precision than the native Python float: "
    #                   f"Tx is {lon.dtype.name}, and Ty is {lat.dtype.name}. "
    #                   "To minimize precision loss, you may want to cast the values to "
    #                   "`float` or `numpy.float64` via the NumPy method `.astype()`.")
    #
    #     # Calculate the distance to the surface of the Sun using the law of cosines
    #     cos_alpha = np.cos(lat) * np.cos(lon)
    #     c = self.observer.radius ** 2 - self.rsun ** 2
    #     b = -2 * self.observer.radius * cos_alpha
    #     # Ignore sqrt of NaNs
    #     with np.errstate(invalid='ignore'):
    #         d = ((-1 * b) - np.sqrt(b ** 2 - 4 * c)) / 2  # use the "near" solution
    #
    #     if self._spherical_screen:
    #         sphere_center = self._spherical_screen['center'].transform_to(self).cartesian
    #         c = sphere_center.norm() ** 2 - self._spherical_screen['radius'] ** 2
    #         b = -2 * sphere_center.dot(rep)
    #         # Ignore sqrt of NaNs
    #         with np.errstate(invalid='ignore'):
    #             dd = ((-1 * b) + np.sqrt(b ** 2 - 4 * c)) / 2  # use the "far" solution
    #
    #         d = np.fmin(d, dd) if self._spherical_screen['only_off_disk'] else dd
    #
    #     # This warning can be triggered in specific draw calls when plt.show() is called
    #     # we can not easily prevent this, so we check the specific function is being called
    #     # within the stack trace.
    #     stack_trace = traceback.format_stack()
    #     matching_string = 'wcsaxes.*_draw_grid'
    #     bypass = any([re.search(matching_string, string) for string in stack_trace])
    #     if not bypass and np.all(np.isnan(d)) and np.any(np.isfinite(cos_alpha)):
    #         warn_user("The conversion of these 2D helioprojective coordinates to 3D is all NaNs "
    #                   "because off-disk coordinates need an additional assumption to be mapped to "
    #                   "calculate distance from the observer. Consider using the context manager "
    #                   "`Helioprojective.assume_spherical_screen()`.")
    #
    #     return self.realize_frame(SphericalRepresentation(lon=lon,
    #                                                       lat=lat,
    #                                                       distance=d))


def _get_rotation_matrix_and_position(obstime):
    r"""
    Return rotation matrix STIX Imaging to SOLO HPC and position of SOLO in HEEQ.

    Parameters
    ----------
    obstime : `astropy.time.Time`
        Time
    Returns
    -------
    """
    roll, pitch, yaw, sas_x, sas_y, solo_position_heeq = _get_aux_data(obstime)
    if np.isnan(sas_x) or np.isnan(sas_y):
        warnings.warn('SAS solution not available using spacecraft pointing and average SAS offset')
        sas_x = yaw
        sas_y = pitch

    # Generate the rotation matrix using the x-convention (see Goldstein)
    axes_matrix = np.array([[1, 0, 0],
                            [0, 0, 1],
                            [0, 1, 0]])
    C = rotation_matrix(sas_x + STIX_X_OFFSET, "z")
    B = rotation_matrix(-1*sas_y + STIX_Y_OFFSET, "y")
    A = rotation_matrix(-1*roll, 'x')
    rmatrix = matrix_product(A, B, C, axes_matrix)

    return rmatrix, solo_position_heeq


def _get_aux_data(obstime):
    # Find, download, read aux file with pointing, sas and position information
    logger.debug('Searching for AUX data')
    query = Fido.search(a.Time(obstime, obstime), a.Instrument.stix, a.Level.l2,
                        a.stix.DataType.aux, a.stix.DataProduct.aux_auxiliary)
    if len(query['stix']) != 1:
        raise ValueError('Exactly one AUX file should be found but %d were found.',
                         len(query['stix']))
    logger.debug('Downloading AUX data')
    files = Fido.fetch(query)
    if len(files.errors) > 0:
        raise ValueError('There were errors downloading the data.')
    # Read and extract data
    logger.debug('Loading and extracting AUX data')
    hdu = fits.getheader(files[0], ext=0)
    aux = QTable.read(files[0], hdu=2)
    start_time = Time(hdu.get('DATE-BEG'))
    rel_date = (obstime - start_time).to('s')
    idx = np.argmin(np.abs(rel_date - aux['time']))
    sas_x, sas_y = aux[idx][['y_srf', 'z_srf']]
    roll, yaw, pitch = aux[idx]['roll_angle_rpy']
    solo_position_heeq = aux[idx]['solo_loc_heeq_zxy']
    logger.debug('SAS: %s, %s, Orientation: %s, %s, %s, Position: %s', sas_x, sas_y, roll, yaw,
                 pitch, solo_position_heeq)
    return roll, pitch, yaw, sas_x, sas_y, solo_position_heeq


@frame_transform_graph.transform(coord.FunctionTransform, STIXImagingFrame, Helioprojective)
def stixim_to_hpc(stxcoord, hpcframe):
    r"""
    Transform STIX Imaging coordinates to given HPC frame
    """
    logger.debug('STIX: %s', stxcoord)

    rot_matrix, solo_pos_heeq = _get_rotation_matrix_and_position(stxcoord.obstime)
    solo_heeq = HeliographicStonyhurst(solo_pos_heeq, representation_type='cartesian',
                                       obstime=stxcoord.obstime)

    # Transform from STIX imaging to SOLO HPC
    newrepr = stxcoord.cartesian.transform(rot_matrix)

    # Create SOLO HPC
    solo_hpc = Helioprojective(newrepr, obstime=stxcoord.obstime,
                               observer=solo_heeq.transform_to(HeliographicStonyhurst))
    logger.debug('SOLO HPC: %s', solo_hpc)

    # Transform from SOLO HPC to input HPC
    hpc = solo_hpc.transform_to(hpcframe)
    logger.debug('Target HPC: %s', hpc)
    return hpc


@frame_transform_graph.transform(coord.FunctionTransform, Helioprojective, STIXImagingFrame)
def hpc_to_stixim(hpccoord, stxframe):
    r"""
    Transform HPC coordinate to STIX Imaging frame.
    """
    logger.debug('Input HPC: %s', hpccoord)
    rmatrix, solo_pos_heeq = _get_rotation_matrix_and_position(stxframe.obstime)
    solo_heeq = HeliographicStonyhurst(solo_pos_heeq, representation_type='cartesian',
                                       obstime=stxframe.obstime)

    # Create SOLO HPC
    solo_hpc_frame = Helioprojective(obstime=stxframe.obstime,
                                     observer=solo_heeq.transform_to(HeliographicStonyhurst))
    # Transform input to SOLO HPC
    solo_hpc_coord = hpccoord.transform_to(solo_hpc_frame)
    logger.debug('SOLO HPC: %s', solo_hpc_coord)

    # Transform from SOLO HPC to STIX imaging
    newrepr = solo_hpc_coord.cartesian.transform(matrix_transpose(rmatrix))
    stx_coord = stxframe.realize_frame(newrepr, obstime=solo_hpc_frame.obstime)
    logger.debug('STIX: %s', stx_coord)
    return stx_coord



from sunpy.net.attr import SimpleAttr

__all__ = ['DataType', 'DataProduct']

DATA_TYPES = ['QL', 'SCI', 'HK', 'CAL']
DATA_PRODUCTS = ['ql_lightcurve', 'ql_background', 'ql_variance', 'ql_spectra',
                 'ql_calibration_spectrum', 'ql_flareflag', 'sci_xray_l0','sci_xray_l1',
                 'sci_xray_l2', 'sci_xray_l3', 'sci_spectrogram', 'sci_visibility']


class DataType(SimpleAttr):
    """
    Data type Quicklook (QL), science (SCI), or housekeeping (HK)
    """
    # def __init__(self, value):
    #     if str(value).casefold not in [t.casefold for t in DATA_TYPES]:
    #         raise ValueError(f'Data type must be one of {DATA_TYPES} not {value}')
    #     super().__init__(value)


class DataProduct(SimpleAttr):
    """
    Data product
    """
    # def __init__(self, value):
    #     if str(value).casefold not in [t.casefold for t in DATA_PRODUCTS]:
    #         raise ValueError(f'Data product must be one of {DATA_PRODUCTS} not {value}')
    #     super().__init__(value)



from datetime import datetime

import pytest

from stixcore.calibration.energy import correct_counts, get_elut
from stixcore.products.product import Product


def test_get_elut():
    with pytest.raises(ValueError, match=r"No ELUT for for date.*"):
        get_elut(datetime(2019, 1, 1))

    elut = get_elut(datetime(2020, 6, 7))
    assert elut.file == 'elut_table_20200519.csv'

    elut = get_elut(datetime.now())
    assert elut.file == 'elut_table_20220601.csv'


@pytest.mark.skip(reason="Better test data")
def test_correct_counts():
    uncorrected_prod = Product('/Users/shane/Downloads/solo_L1_stix-sci-xray-cpd-'
                               '2109270021_20210927T085625_20210927T092350_V01_63392.fits')
    corrected_prod = correct_counts(uncorrected_prod)
    assert corrected_prod.e_cal == 'rebin'



from pathlib import Path

import numpy as np

from stixcore.io.elut import read_elut, read_elut_index
from stixcore.products.common import rebin_proportional
from stixcore.products.level0.scienceL0 import ScienceProduct


def get_elut(date):
    root = Path(__file__).parent.parent
    elut_index_file = Path(root, *['config', 'data', 'common', 'elut', 'elut_index.csv'])

    elut_index = read_elut_index(elut_index_file)
    elut_info = elut_index.at(date)
    if len(elut_info) == 0:
        raise ValueError(f'No ELUT for for date {date}')
    elif len(elut_info) > 1:
        raise ValueError(f'Multiple ELUTs for for date {date}')
    start_date, end_date, elut_file = list(elut_info)[0]
    elut_table = read_elut(elut_file)

    return elut_table


def correct_counts(product, method='rebin'):
    """
    Correct count for individual pixel energy calibration


    Parameters
    ----------
    product :
        A stix product
    method : str optional
        The correction method default is to rebin onto science energy channels

    Returns
    -------
    The product with the correct counts
    """
    if method not in ['rebin', 'width']:
        raise ValueError("method not on of the supported methods 'rebin' or 'width'")

    if not isinstance(product, ScienceProduct) and product.level == 'L1':
        raise ValueError('Only supports L1 science products')

    elut = get_elut(product.utc_timerange.center.datetime)

    # rebin back onto science channels
    counts = product.data['counts'][...]
    for did, pid in zip(elut.detector.flat, elut.pixel.flat):
        sum_orig_counts = counts[:, did, pid, 1:-1].sum(axis=-1)
        calibration_edges = elut.e_actual[did, pid, :]

        if method == 'rebin':
            # need bin below 4 and above 150 to conserve total counts so [0, 4, ... 150, 160]
            sci_edges = np.hstack([elut.e, 160])
            rebinned = np.apply_along_axis(rebin_proportional, -1, counts[:, did, pid, 1:-1],
                                           calibration_edges, sci_edges)

            sum_rebined_counts = rebinned.sum(axis=-1)
            if not np.allclose(sum_orig_counts.value, sum_rebined_counts):
                raise ValueError('We are losing the precious counts')

            # zero the counts the were rebinned then add all 32 ch
            counts[:, did, pid, 1:-1] = 0
            counts[:, did, pid, :] = counts[:, did, pid, :] + rebinned * counts.unit
        elif method == 'width':
            real_width = np.diff(calibration_edges)
            counts[:, pid, did, 1:-1] = counts[:, pid, did, 1:-1]/real_width

    product.data['counts'] = counts
    product.e_cal = method

    return product



from pathlib import Path

import pytest
from sunpy.timeseries import TimeSeries

from stixpy.data import test
from stixpy.timeseries.quicklook import *


def test_ql_lightcurve():
    ql_lc = TimeSeries(test.STIX_QL_LIGHTCURVE_TIMESERIES)
    assert isinstance(ql_lc, QLLightCurve)


def test_qlbackground():
    ql_lc = TimeSeries(test.STIX_QL_BACKGROUND_TIMESERIES)
    assert isinstance(ql_lc, QLBackground)


def test_qlvariance():
    ql_lc = TimeSeries(test.STIX_QL_VARIANCE_TIMESERIES)
    ql_lc.peek()
    assert isinstance(ql_lc, QLVariance)



from pathlib import Path
from functools import partial
from collections import OrderedDict

import numpy as np
from roentgen.absorption.material import Compound, MassAttenuationCoefficient, Material

import astropy.units as u
from astropy.table.table import Table

from stixcore.config.reader import read_energy_channels

__all__ = ['Transmission']

MIL_SI = 0.0254 * u.mm

# TODO move to configuration files
COMPONENTS = OrderedDict([
    ('front_window', [('solarblack', 0.005 * u.mm), ('be_alloy', 2*u.mm)]),
    ('rear_window', [('be_alloy', 1*u.mm)]),
    ('grid_covers', [('kapton', 4 * 2 * MIL_SI)]),
    ('dem', [('kapton', 2 * 3 * MIL_SI)]),
    ('attenuator', [('al_alloy', 0.6 * u.mm)]),
    ('mli', [('al', 1000 * u.angstrom), ('kapton', 3 * MIL_SI), ('al', 40 * 1000 * u.angstrom),
             ('mylar', 20 * 0.25 * MIL_SI), ('pet', 21 * 0.005 * u.mm), ('kapton', 3 * MIL_SI),
             ('al', 1000 * u.angstrom)]),
    ('calibration_foil', [('al', 4 * 1000 * u.angstrom), ('kapton', 4 * 2 * MIL_SI)]),
    ('dead_layer', [('te_o2', 392 * u.nm)]),
])

# For attenuator see https://www.metallservice.ch/msm/msm-home/services/infoservice/
# produktinfos-datenbl%C3%A4tter/aluminium-platten/en_aw-7075_1-3.pdf

MATERIALS = OrderedDict([
    ('al', ({'Al': 1.0}, 2.7 * u.g/u.cm**3)),
    ('al_alloy', ({'Al': 0.89345, 'Si': 0.002, 'Fe': 0.0025, 'Cu': 0.016, 'Mn': 0.0015, 'Mg': 0.025,
                   'Cr': 0.0023, 'Ni': 0.00025, 'Zn': 0.056, 'Ti': 0.001}, 2.8 * u.g/u.cm**3)),
    ('be_alloy', ({'Al': 0.0005, 'Be': 0.9974, 'C': 0.00075, 'Fe': 0.00065, 'Mg': 0.0004,
                   'Si': 0.0003}, 1.84 * u.g/u.cm**3)),
    ('kapton', ({'H': 0.026362, 'C': 0.691133, 'N': 0.073270, 'O': 0.209235},
                1.43 * u.g / u.cm ** 3)),
    ('mylar', ({'H': 0.041959, 'C': 0.625017, 'O': 0.333025}, 1.38 * u.g / u.cm ** 3)),
    ('pet', ({'H': 0.041960, 'C': 0.625016, 'O': 0.333024}, 1.370 * u.g / u.cm**3)),
    ('solarblack_oxygen', ({'H': 0.002, 'O': 0.415, 'Ca': 0.396, 'P': 0.187}, 3.2 * u.g/u.cm**3)),
    ('solarblack_carbon', ({'C': 0.301, 'Ca': 0.503, 'P': 0.195}, 3.2 * u.g/u.cm**3)),
    ('te_o2', ({'Te': 0.7995088158691722, 'O': 0.20049124678825841}, 5.670 * u.g/u.cm**3))]
)

# TODO get file from config
ENERGY_CHANNELS = read_energy_channels(Path(__file__).parent.parent / "config" / "data" /
                                       "common" / "detector" / "ScienceEnergyChannels_1000.csv")


class Transmission:
    """
    Calculate the energy dependant transmission of X-ray through the instrument
    """
    def __init__(self, solarblack='solarblack_carbon'):
        """
        Create a new instance of the transmission with the given solar black composition.

        Parameters
        ----------

        solarblack : `str` optional
            The SolarBlack composition to use.
        """
        if solarblack not in ['solarblack_oxygen', 'solarblack_carbon']:
            raise ValueError("solarblack must be either 'solarblack_oxygen' or "
                             "'solarblack_carbon'.")

        self.solarblack = solarblack
        self.materials = MATERIALS
        self.components = COMPONENTS
        self.components = dict()
        self.energies = [ENERGY_CHANNELS[i].e_lower for i in range(1, 32)] * u.keV

        for name, layers in COMPONENTS.items():
            parts = []
            for material, thickness in layers:
                if material == 'solarblack':
                    material = self.solarblack
                mass_frac, den = MATERIALS[material]
                parts.append(self.create_material(name=material, fractional_masses=mass_frac,
                                                  thickness=thickness, density=den))
            self.components[name] = Compound(parts)

    def get_transmission(self, energies=None, attenuator=False):
        """
        Get the transmission for each detector at the center of the given energy bins.

        If energies are not supplied will evaluate at standard science energy channels

        Parameters
        ----------
        energies : `astropy.units.Quantity`, optional
            The energies to evaluate the transmission
        attenuator : `bool`, optional
            True for attenuator in X-ray path, False for attenuator not in X-ray path

        Returns
        -------
        `astropy.table.Table`
            Table containing the transmission values for each energy and detector
        """
        base_comps = [self.components[name] for name in ['front_window', 'rear_window', 'dem',
                                                         'mli', 'calibration_foil',
                                                         'dead_layer']]

        if energies is None:
            energies = self.energies

        if attenuator:
            base_comps.append(self.components['attenuator'])

        base = Compound(base_comps)
        base_trans = base.transmission(energies)

        fine = Compound(base_comps + [self.components['grid_covers']])
        fine_trans = fine.transmission(energies)

        # TODO need to move to configuration db
        fine_grids = np.array([11, 13, 18, 12, 19, 17]) - 1
        transmission = Table()
        # transmission['sci_channel'] = range(1, 31)
        transmission['energies'] = energies
        for i in range(32):
            name = f'det-{i}'
            if np.isin(i, fine_grids):
                transmission[name] = fine_trans
            else:
                transmission[name] = base_trans
        transmission['attenuator'] = self.components['attenuator'].transmission(energies)
        return transmission

    def get_transmission_by_component(self):
        """
        Get the contributions to the total transmission by broken down by component.

        Returns
        -------
        `dict`
            Entries are Compounds for each component
        """
        return self.components

    def get_transmission_by_material(self):
        """
        Get the contribution to the transmission by total thickness for each material.

        Layers of the same materials are combined to return one instance with the total thickness.

        Returns
        -------
        `dict`
            Entries are meterials with the total thickness for that material.
        """
        material_thickness = dict()
        for name, layers in COMPONENTS.items():
            for material_name, thickness in layers:
                if material_name == 'solarblack':
                    material_name = self.solarblack
                if material_name in material_thickness.keys():
                    material_thickness[material_name] += thickness.to('mm')
                else:
                    material_thickness[material_name] = thickness.to('mm')
        res = {}
        for name, thickness in material_thickness.items():
            frac_mass, density = self.materials[name]
            mat = self.create_material(name=name, fractional_masses=frac_mass,
                                       density=density, thickness=thickness)
            res[name] = mat

        return res

    @classmethod
    def create_material(cls, name=None, fractional_masses=None, thickness=None, density=None):
        """
        Create a new material given the composition and fractional masses.

        Parameters
        ----------
        name : `str`
            Name of the meterial
        fractional_masses : `dict`
            The element and fractional masses of the material e.g. `{'H': 0.031, 'O': 0.969}`
        thickness : `astropy.units.Quantity`
            Thickness of the material
        density : `astropy.units.Quantity`
            Density of the material

        Returns
        -------
        `roentgen.absorption.material.Material`
            The material
        """
        material = Material('h', thickness, density)
        material.name = name
        # probably don't need this
        material.density = density

        # TODO remove in favour of upstream fix when completed
        #  see https://github.com/ehsteve/roentgen/issues/26
        def func(composition, e):
            return sum([MassAttenuationCoefficient(element).func(e) * frac_mass
                        for element, frac_mass in composition.items()])

        material_func = partial(func, fractional_masses)

        material.mass_attenuation_coefficient.func = material_func
        return material


def generate_transmission_tables():
    from datetime import datetime
    cur_date = datetime.now().strftime('%Y%m%d')
    datetime.now().strftime('%Y%m%d')
    trans = Transmission()

    energies = np.linspace(2, 150, 1001) * u.keV

    norm_sci_energies = trans.get_transmission()
    norm_sci_energies.write(f'stix_transmission_sci_energies_{cur_date}.csv')
    norm_high_res = trans.get_transmission(energies=energies)
    norm_high_res.write(f'stix_transmission_highres_{cur_date}.csv')

    comps = trans.get_transmission_by_component()

    comps_sci_energies = Table([c.transmission(trans.energies) for c in comps.values()],
                               names=[k for k in comps.keys()])
    comps_sci_energies['energy'] = trans.energies
    comps_sci_energies.write(f'stix_transmission_by_component_sci_energies_{cur_date}.csv')

    comps_highres = Table([c.transmission(energies) for c in comps.values()],
                          names=[k for k in comps.keys()])
    comps_highres['energy'] = energies
    comps_highres.write(f'stix_transmission_by_component_highres_{cur_date}.csv')



# Note that we need to fall back to the hard-coded version if either
# setuptools_scm can't be imported or setuptools_scm can't determine the
# version, so we catch the generic 'Exception'.
try:
    from setuptools_scm import get_version
    __version__ = get_version(root='..', relative_to=__file__)
except Exception:
    __version__ = '0.1.0rc1'



"""
STIX integer compression is a quasi-exponential 1 byte compression of integer data, both positive
or negative depending on the parameterization.  For the decompression it returns the mid-point value
of the possible compressed values that could have given the input.

The compression algorithm is applied to an integer value, V, to yield a compressed value, C, as
follows:

1. If  V < 0, set s=1 and set V = -V
2. If  V < 2^(M+1), set C = V and skip to step 6
3. If  V >= 2^(M+1), shift V right until 1s (if any) appear only in LS M+1 bits
4. Exponent, e = number of shifts+ 1
5. Mantissa, m = LS M bits of the shifted value.
6. Set C = m + 2^M * e
7. If S=1, set msb of C = s

The algorithm family has the following properties:

* The compressed value is exact for input absolute values up to 2^(M+1)
* RMS fractional error is in range: 1./(sqrt(12)*2^M) to 1/(sqrt(1/12)*s^(M+1))
* The maximum absolute value that can be compressed is given by: 2^(2^K-2) x (2^(M+1)-1)
  The number of bits required to hold the compressed value is S+K+M

References
----------
STIX-TN-0117-FHNW
"""
import numpy as np

__all__ = ['compress', 'decompress', 'CompressionRangeError', 'CompressionSchemeParameterError',
           'NonIntegerCompressionError']


def compress(values, *, s, k, m):
    """
    Compress values according to parameters.

    Parameters
    ----------
    values : array-like (int)
        Values to be compressed
    s : int
        Number of sign bits, 0 for positive only values 1 for signed
    k : int
        Number of bits for exponent
    m : int
        Number of bit for mantissa

    Returns
    -------
    array
        The compressed values

    Examples
    --------
    >>> comp = compress(12345, s=0, k=5, m=3)
    >>> int(comp)
    92
    >>> comp = compress(-1984, s=1, k=3, m=4)
    >>> int(comp)
    255
    """
    values = np.atleast_1d(np.array(values))
    if not np.issubdtype(values.dtype, np.integer):
        raise NonIntegerCompressionError(f'Input must be an integer type not {values.dtype}')

    total = s + k + m
    if total > 8:
        raise CompressionSchemeParameterError(f'Invalid s={s}, k={k}, m={m} '
                                              f'must sum to less than 8 not {total}')

    max_value = 2**(2**k-2)*(2**(m + 1)) - 1
    abs_range = [0, max_value] if s == 0 else [-1 * max_value, 1 * max_value]

    if values.min() < abs_range[0] or values.max() > abs_range[1]:
        raise CompressionRangeError(f'Valid input range exceeded for s={s}, k={k}, m={m} '
                                    f'input (min/max) {values.min()}/{values.max()} '
                                    f'exceeds {abs_range}')

    negative_indices = np.nonzero(values < 0)
    out = np.abs(values).astype(np.uint64)
    compress_indices = np.nonzero(out >= 2 ** (m + 1))

    if values[compress_indices].size != 0:
        kv = np.floor((np.log(np.abs(values[compress_indices])) / np.log(2.0) - m)).astype(np.int64)
        mv = (np.abs(values[compress_indices], dtype=np.int64) // 2**kv) & (2**m - 1)
        out[compress_indices] = kv + 1 << m | mv

    if s != 0 and values[negative_indices].size >= 1:
        out[negative_indices] += 128
    return out


def decompress(values, *, s, k, m, return_variance=False):
    """
    Compress values according to parameters.

    Parameters
    ----------
    values : array-like (int)
        Values to decompressed
    s : int (0, 1)
        Number of sign bits
    k : int
        Number of bits for exponent
    m : int
        Number of bits for mantissa
    return_variance : boolean (optional)

    Returns
    -------
    array
        The decompressed values

    Examples
    --------
    >>> decomp = decompress(92, s=0, k=5, m=3)
    >>> int(decomp)
    12799
    >>> decomp = decompress(255, s=1, k=3, m=4)
    >>> int(decomp)
    -2015
    """
    values = np.atleast_1d(np.array(values))
    if not np.issubdtype(values.dtype, np.integer):
        raise NonIntegerCompressionError(f'Input must be an integer type not {values.dtype}')

    values = values.astype(np.uint64)

    total = s + k + m
    if total > 8:
        raise CompressionSchemeParameterError(f'Invalid s={s}, k={k}, m={m} '
                                              f'must sum to less than 8 not {total}')

    if values.min() < 0 or values.max() > 255:
        raise CompressionRangeError(f'Compressed values must be in the range 0 to 255')

    max_value = 2**(2**k-2)*(2**(m + 1)) - 1
    uint64_info = np.iinfo(np.uint64)

    if max_value > uint64_info.max:
        raise CompressionRangeError('Decompressed value too large to fit into uint64')

    abs_values = values if s == 0 else values & 127
    out = abs_values[:].astype(np.uint64)
    negative_indices = np.nonzero(values >= 128) if s != 0 else ([])
    decompress_indices = np.nonzero(abs_values >= 2**(m+1))

    if return_variance:
        variance = np.zeros_like(values, dtype=np.float64)

    if abs_values[decompress_indices].size >= 1:
        kv = (abs_values[decompress_indices] >> m) - 1
        mv = 2**m + (abs_values[decompress_indices] & 2**m - 1)
        out[decompress_indices] = (mv << kv) + (1 << (kv - 1)) - 1
        if return_variance:
            variance[decompress_indices] = ((1 << kv)**2 + 2)/12

    if abs_values[negative_indices].size >= 1:
        out = out.astype(np.int64)
        out[negative_indices] = -1 * out[negative_indices]

    if return_variance:
        return out, variance
    else:
        return out


class NonIntegerCompressionError(Exception):
    """
    Exception raised when the input to the compression algorithm is not an integer.
    """


class CompressionSchemeParameterError(Exception):
    """
    Exception raised for invalid compression scheme parameters.
    """


class CompressionRangeError(Exception):
    """
    Exception raised due to invalid range in the input or output.
    """
